% learning_clafic.m 学習部分空間法
clear
close all

r=1;     % 部分空間の次元数(各クラス共通, 1<=r<=9)
alp=0.1; % 学習係数 (0<alp<1)

load('traiX.mat');    % traiのオリジナル画像とラベルの読込み
load('queryX.mat');   % queryのオリジナル画像とラベルの読込み

nclass=40; % クラス数
[d ntrai] =size(trai.X);   % 画像の次元数dと訓練画像数ntrai
[d nquery]=size(query.X);  % 画像の次元数dとテスト画像数nquery

% 部分空間の直交系の算出
for jj = 1 : nclass
	% ラベルがjjの訓練画像をYに格納
	X=trai.X(:,find(trai.label==jj));
	[d n] = size(X);           % Xのサイズ取得 n<<d
	N = X'*X;                  % n x nの小さい行列Nを計算
	rdim = rank(N);            % Nの階数を算出

	% 固有ベクトルを並べた行列Uの算出
	[eig_vec eig_val] = eig(N); % Nの固有値と固有ベクトルを求める
	[value index]=sort(diag(eig_val),'descend'); % 固有値を降順に並べ替え
	U = X * eig_vec(:,index(1:rdim)) * diag(sqrt(1./value(1:rdim)) );

	C(jj).U  = U; % クラスjjの固有ベクトル
	fprintf(2,'class %d --- ok\n',jj);
end

%% 学習
s=zeros(nclass,1);         % 類似度格納用変数
MAX_ITE=100;               % 最大学習回数
obj_func=zeros(MAX_ITE,1); % 学習ステップごとの識別率

% Uの初期化
for jj = 1 : nclass
	LC(jj).U=C(jj).U(:,1:r);
end

for t=1:MAX_ITE

	% 修正量用変数の初期化
	for jj = 1 : nclass
		dC(jj).U=zeros(d,r);
  end

	% 訓練画像を用いて学習
  for ii = 1 : ntrai
  	
  	% 各クラスとの類似度計算
  	for jj = 1 : nclass
  		s(jj)=norm((LC(jj).U)'*trai.X(:,ii)).^2;
  	end

  	% 類似度を降順にソート  	
  	[value index]=sort(s,'descend');
  	
  	% 訓練画像のラベルをppに格納
  	pp=trai.label(ii);
  	
  	% 訓練画像のラベルをppに格納
  	if index(1)~=trai.label(ii)
  		nn=index(1);  % 誤識別の場合はそのラベルをnnに
  	else
  		nn=index(2);  % 正識別の場合は二番目のクラスのラベルをnnに
    end
    
    % 係数の計算
    s1=s(pp);, s2=s(nn);
    u=(s2-s1)./(s1+s2);
    f=1./(1+exp(-u.*t));, df=alp.*t.*f.*(1-f);
    ds1=s2./(s1+s2).^2;, ds2=s1./(s1+s2).^2;
		
		% 評価関数値の計算
		obj_func(t)=obj_func(t)+f;

    % 修正量の計算
    dC(pp).U=dC(pp).U+df.*ds1.*(trai.X(:,ii)*((trai.X(:,ii))'*LC(pp).U));
    dC(nn).U=dC(nn).U-df.*ds2.*(trai.X(:,ii)*((trai.X(:,ii))'*LC(nn).U));
	end
	
  fprintf(2,'learning step t = %3d, object function value =%f\n',t,obj_func(t));
	
	% 評価関数の変化が小さくなったら終了
	if t>1
		if abs(obj_func(t-1)-obj_func(t))<1E-5
			break;
		else
			% 正規直交化
			for jj = 1 : nclass
				LC(jj).U=LC(jj).U+dC(jj).U;
				LC(jj).U=orth(LC(jj).U);
			end
		end
	else
		for jj = 1 : nclass
			LC(jj).U=LC(jj).U+dC(jj).U;
			LC(jj).U=orth(LC(jj).U);
		end
	end
end

% 学習後のテスト画像の識別率
accuracy=0;             % 識別率
for ii = 1 : nquery
	% 各クラスの類似度を計算
	for jj = 1 : nclass
		s(jj)=norm((LC(jj).U)'*query.X(:,ii));
	end
	% 類似度が最大となるクラスを求める
	[value index]=max(s);
	% 類似度が最大となるクラスとテストデータのラベルが等しい
	if index==query.label(ii)
		accuracy=accuracy+1;
	end
end
% 識別率を百分率で表現
accuracy=(accuracy./nquery).*100;
fprintf('dimensionality r=%d, accuracy : %2.1f \n',r,accuracy);
figure(1),clf,plot(obj_func(1:t));
xlabel('learning step t');, ylabel('Object function value');
